using Adaptation.Shared;
using Adaptation.Shared.Methods;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.Json;

namespace Adaptation.FileHandlers.json;

public class Description : IDescription, Shared.Properties.IDescription
{

    public int Test { get; set; }
    public int Count { get; set; }
    public int Index { get; set; }
    //
    public string EventName { get; set; }
    public string NullData { get; set; }
    public string JobID { get; set; }
    public string Sequence { get; set; }
    public string MesEntity { get; set; }
    public string ReportFullPath { get; set; }
    public string ProcessJobID { get; set; }
    public string MID { get; set; }
    //
    public string Date { get; set; }
    public string Employee { get; set; }
    public string HeaderUniqueId { get; set; }
    public string Layer { get; set; }
    public string PSN { get; set; }
    public string RDS { get; set; }
    public string Reactor { get; set; }
    public string Recipe { get; set; }
    public string UniqueId { get; set; }
    public string Zone { get; set; }
    //
    public string SineBevelAngle { get; set; }
    public string XStep { get; set; }
    //
    public string Number { get; set; }
    public string Depth { get; set; }
    public string Raw { get; set; }
    public string Edited { get; set; }
    public string Resistivity { get; set; }
    public string CD { get; set; }

    string IDescription.GetEventDescription() => "File Has been read and parsed";

    List<string> IDescription.GetNames(IFileRead fileRead, Logistics logistics)
    {
        List<string> results = new();
        IDescription description = GetDefault(fileRead, logistics);
        string json = JsonSerializer.Serialize(description, description.GetType());
        object @object = JsonSerializer.Deserialize<object>(json);
        if (@object is not JsonElement jsonElement)
            throw new Exception();
        foreach (JsonProperty jsonProperty in jsonElement.EnumerateObject())
            results.Add(jsonProperty.Name);
        return results;
    }

    List<string> IDescription.GetDetailNames()
    {
        List<string> results = new()
        {
            nameof(HeaderUniqueId),
            nameof(Layer),
            nameof(RDS),
            nameof(UniqueId),
            nameof(Zone)
        };
        return results;
    }

    List<string> IDescription.GetHeaderNames()
    {
        List<string> results = new()
        {
            nameof(Date),
            nameof(Employee),
            nameof(PSN),
            nameof(RDS),
            nameof(Reactor),
            nameof(Recipe)
        };
        return results;
    }

    IDescription IDescription.GetDisplayNames()
    {
        Description result = GetDisplayNames();
        return result;
    }

    List<string> IDescription.GetParameterNames()
    {
        List<string> results = new()
        { };
        return results;
    }

    JsonProperty[] IDescription.GetDefault(IFileRead fileRead, Logistics logistics)
    {
        JsonProperty[] results;
        IDescription description = GetDefault(fileRead, logistics);
        string json = JsonSerializer.Serialize(description, description.GetType());
        object @object = JsonSerializer.Deserialize<object>(json);
        results = ((JsonElement)@object).EnumerateObject().ToArray();
        return results;
    }

    List<string> IDescription.GetPairedParameterNames()
    {
        List<string> results = new();
        return results;
    }

    List<string> IDescription.GetIgnoreParameterNames(Test test)
    {
        List<string> results = new();
        return results;
    }

    IDescription IDescription.GetDefaultDescription(IFileRead fileRead, Logistics logistics)
    {
        Description result = GetDefault(fileRead, logistics);
        return result;
    }

    Dictionary<string, string> IDescription.GetDisplayNamesJsonElement(IFileRead fileRead)
    {
        Dictionary<string, string> results = new();
        IDescription description = GetDisplayNames();
        string json = JsonSerializer.Serialize(description, description.GetType());
        JsonElement jsonElement = JsonSerializer.Deserialize<JsonElement>(json);
        foreach (JsonProperty jsonProperty in jsonElement.EnumerateObject())
        {
            if (!results.ContainsKey(jsonProperty.Name))
                results.Add(jsonProperty.Name, string.Empty);
            if (jsonProperty.Value is JsonElement jsonPropertyValue)
                results[jsonProperty.Name] = jsonPropertyValue.ToString();
        }
        return results;
    }

    List<IDescription> IDescription.GetDescriptions(IFileRead fileRead, Logistics logistics, List<Test> tests, IProcessData iProcessData)
    {
        List<IDescription> results = new();
        if (iProcessData is null || !iProcessData.Details.Any() || iProcessData is not ProcessData processData)
            results.Add(GetDefault(fileRead, logistics));
        else
        {
            string nullData;
            Description description;
            object configDataNullData = fileRead.NullData;
            if (configDataNullData is null)
                nullData = string.Empty;
            else
                nullData = configDataNullData.ToString();
            for (int i = 0; i < iProcessData.Details.Count; i++)
            {
                if (iProcessData.Details[i] is not Detail detail)
                    continue;
                description = new Description
                {
                    Test = (int)tests[i],
                    Count = tests.Count,
                    Index = i,
                    //
                    EventName = fileRead.EventName,
                    NullData = nullData,
                    JobID = fileRead.CellInstanceName,
                    Sequence = logistics.Sequence.ToString(),
                    MesEntity = logistics.MesEntity,
                    ReportFullPath = logistics.ReportFullPath,
                    ProcessJobID = logistics.ProcessJobID,
                    MID = logistics.MID,
                    //
                    Date = processData.Date.ToString(GetDateFormat()),
                    Employee = processData.Employee,
                    Layer = processData.Layer,
                    PSN = processData.PSN,
                    RDS = processData.RDS,
                    Reactor = processData.Reactor,
                    Recipe = processData.Recipe,
                    Zone = processData.Zone,
                    //
                    SineBevelAngle = processData.SineBevelAngle,
                    XStep = processData.XStep,
                    //
                    UniqueId = detail.UniqueId,
                    HeaderUniqueId = detail.HeaderUniqueId,
                    Number = detail.Number,
                    Depth = detail.Depth,
                    Raw = detail.Raw,
                    Edited = detail.Edited,
                    Resistivity = detail.Resistivity,
                    CD = detail.CD,
                };
                results.Add(description);
            }
        }
        return results;
    }

    private static Description GetDisplayNames()
    {
        Description result = new();
        return result;
    }

    private Description GetDefault(IFileRead fileRead, Logistics logistics)
    {
        Description result = new()
        {
            Test = -1,
            Count = 0,
            Index = -1,
            //
            EventName = fileRead.EventName,
            NullData = fileRead.NullData,
            JobID = fileRead.CellInstanceName,
            Sequence = logistics.Sequence.ToString(),
            MesEntity = fileRead.MesEntity,
            ReportFullPath = logistics.ReportFullPath,
            ProcessJobID = logistics.ProcessJobID,
            MID = logistics.MID,
            //
            Date = nameof(Date),
            Employee = nameof(Employee),
            HeaderUniqueId = nameof(HeaderUniqueId),
            Layer = nameof(Layer),
            PSN = nameof(PSN),
            RDS = nameof(RDS),
            Reactor = nameof(Reactor),
            Recipe = nameof(Recipe),
            UniqueId = nameof(UniqueId),
            Zone = nameof(Zone),
            //
            SineBevelAngle = nameof(SineBevelAngle),
            XStep = nameof(XStep),
            //
            Number = nameof(Number),
            Depth = nameof(Depth),
            Raw = nameof(Raw),
            Edited = nameof(Edited),
            Resistivity = nameof(Resistivity),
            CD = nameof(CD),
        };
        return result;
    }

    internal static string GetDateFormat() => "MM/dd/yyyy hh:mm:ss tt";

}